#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Measurement objects in the Bwyd language.
see copyright/license https://github.com/DerwenAI/bwyd/README.md
"""

from collections import OrderedDict
from fractions import Fraction
import enum
import logging
import typing

from icecream import ic  # type: ignore  # pylint: disable=W0611
from pydantic import BaseModel, NonNegativeFloat, PositiveFloat
import inflect


class MeasureUnits (enum.StrEnum):
    """
An enumeration class representing string literals for Measure units.
    """
    TEASPOON = "tsp"
    CUP = enum.auto()
    POUND = enum.auto()
    OUNCE = enum.auto()
    LITER = "l"
    MILLILITER = "ml"
    GRAM = "g"
    KILOGRAM = "kg"

OUNCE_PER_POUND: float = 16.0
CUP_PER_LITER: float = 4.226753
POUND_PER_GRAM: float = 0.002204623


class DurationUnits (enum.StrEnum):
    """
An enumeration class representing string literals for Duration units.
    """
    SECOND = enum.auto()
    MINUTE = enum.auto()
    HOUR = enum.auto()
    DAY = enum.auto()
    MONTH = enum.auto()
    YEAR = enum.auto()


NORM_RATIO: typing.Dict[ str, int ] = OrderedDict({
    DurationUnits.YEAR.value: 60 * 60 * 24 * 365,
    DurationUnits.MONTH.value: 60 * 60 * 24 * 30,
    DurationUnits.DAY.value: 60 * 60 * 24,
    DurationUnits.HOUR.value: 60 * 60,
    DurationUnits.MINUTE.value: 60,
    DurationUnits.SECOND.value: 1,
})


class TemperatureUnits (enum.StrEnum):
    """
An enumeration class representing string literals for Temperature units.
    """
    CELSIUS = "C"
    FAHRENHEIT = "F"


PLURAL = inflect.engine()


class Conversion (BaseModel):  # pylint: disable=R0902
    """
A data class representing one Conversion object.
    """
    symbol: str
    density: PositiveFloat
    imperial: str = MeasureUnits.CUP.value
    metric: str = MeasureUnits.GRAM.value


Converter = typing.Dict[ str, Conversion ]


class Humanized (BaseModel):  # pylint: disable=R0902
    """
A data class representing one humanized Measure object.
    """
    amount: NonNegativeFloat
    human: str
    units: typing.Optional[ MeasureUnits ]


    def denormalize (
        self
        ) -> str:
        """
Denormalize a meaure which is already in human-readable form.
        """
        if self.units is None:
            return f" ({self.human})"

        units: str = self.units.value

        if self.amount > 1.0 and self.human != "1" and self.units != MeasureUnits.TEASPOON:
            units = PLURAL.plural(units)

        return f" ({self.human} {units})"


class Measure (BaseModel):  # pylint: disable=R0902
    """
A data class representing one parsed Measure object.
    """
    amount: NonNegativeFloat
    units: typing.Optional[ str ]


    @classmethod
    def build (
        cls,
        parse: typing.Any,
        ) -> "Measure":
        """
Constructor from a `textx` parse object.
        """
        measure_units: typing.Optional[ str ] = None

        if parse.units is not None:
            measure_units = MeasureUnits(parse.units).value

        return Measure(
            amount = float(parse.amount),
            units = measure_units,
        )


    def humanize (
        self
        ) -> str:
        """
Denormalize this measure into human-readable form.
        """
        html: str = f"{self.amount}"

        if html.endswith(".0"):
            html = html[:-2]

        if self.units is not None:
            html = f"{html} {self.units}"

        return html


    def humanize_convert (
        self,
        symbol: str,
        external: bool,
        converter: typing.Optional[ Converter ],
        ) -> str:
        """
Denormalize this measure into human-readable form, with an
imperial conversion if available.
        """
        amount: str = self.humanize().strip()

        if converter is not None:
            if symbol in converter:
                conv: Conversion = converter[symbol]

                if self.units == conv.metric:
                    imper_amount: float = self.amount / conv.density
                    human: Humanized = self._humanize_cup(imper_amount)
                    amount += human.denormalize()

            elif self.units is not None:
                if not external:
                    logging.warning(f"no conversion ratio for {symbol}")  # pylint: disable=W1203

                match self.units:
                    case MeasureUnits.GRAM.value:
                        imper_amount = self.amount * POUND_PER_GRAM

                        if imper_amount < 0.25:
                            imper_amount *= OUNCE_PER_POUND

                            if imper_amount > 0.95:
                                human = self._humanize_generic(imper_amount, MeasureUnits.OUNCE)
                            else:
                                human = Humanized(
                                    amount = imper_amount,
                                    human = str(round(imper_amount, 2)),
                                    units = MeasureUnits.OUNCE,
                                )
                        else:
                            human = self._humanize_generic(imper_amount, MeasureUnits.POUND)

                        amount += human.denormalize()

                    case MeasureUnits.LITER.value:
                        imper_amount = self.amount * CUP_PER_LITER
                        human = self._humanize_generic(imper_amount, MeasureUnits.CUP)
                        amount += human.denormalize()

                    case _:
                        logging.warning(f"no default conversion for unit `{self.units}`")  # pylint: disable=W1203

        return amount


    @classmethod
    def _humanize_generic (
        cls,
        amount: float,
        units: MeasureUnits,
        ) -> Humanized:
        """
Private method to humanize imperial measurement ratios, for generic case.
        """
        denom_limit: int = 4

        if amount > .95:
            human: str = cls.fix_fraction(amount)
        else:
            human = str(Fraction(round(amount, 2)).limit_denominator(denom_limit))

        return Humanized(
            amount = amount,
            human = human,
            units = units,
        )


    @classmethod
    def _humanize_cup (
        cls,
        amount: float,
        ) -> Humanized:
        """
Private method to humanize imperial measurement ratios, for cup.
        """
        units: MeasureUnits = MeasureUnits.CUP
        denom_limit: int = 4

        if amount <= 0.24:
            return cls._humanize_tsp(amount * 16.0)

        if amount >= 1.0:
            human: str = cls.fix_fraction(amount)
        else:
            human = str(Fraction(round(amount, 2)).limit_denominator(denom_limit))

        return Humanized(
            amount = amount,
            human = human,
            units = units,
        )


    @classmethod
    def _humanize_tsp (
        cls,
        amount: float,
        ) -> Humanized:
        """
Private method to humanize imperial measurement ratios, for teaspoon.
        """
        units: MeasureUnits = MeasureUnits.TEASPOON
        denom_limit: int = 8

        if amount >= 0.95:
            human: str = cls.fix_fraction(amount)
        elif amount >= 0.4:
            human = str(Fraction(round(amount, 1)).limit_denominator(denom_limit))
        else:
            human = str(Fraction(round(amount, 2)).limit_denominator(denom_limit))

        # plural for teaspoons is too easily confused with tablespoon

        return Humanized(
            amount = amount,
            human = human,
            units = units,
        )


    @classmethod
    def fix_fraction (
        cls,
        amount: float,
        ) -> str:
        """
Humanize fractions representing imperial measurement ratios >= 1.0
        """
        base: int = int(amount)
        frac: float = amount - float(base)

        if frac >= 0.9:
            return str(round(amount))
        if frac < 0.2:
            return str(base)

        denom_limit: int = 4
        human = str(Fraction(round(frac, 2)).limit_denominator(denom_limit))

        return f"{base:d} {human}"


    def get_model (
        self
        ) -> dict:
        """
Serializable representation for JSON.
        """
        return {
            "amount": self.amount,
            "units": self.units,
        }


class Duration (Measure):  # pylint: disable=R0902
    """
A data class representing one parsed Duration object.
    """


    @classmethod
    def build (
        cls,
        parse: typing.Any,
        ) -> "Duration":
        """
Constructor from a `textx` parse object.
        """
        duration_units: typing.Optional[ str ] = None

        if parse.units is not None:
            duration_units = DurationUnits(parse.units).value

        return Duration(
            amount = float(parse.amount),
            units = duration_units,
        )


    def normalize (
        self,
        ) -> float:
        """
Return this duration normalized into seconds.
        """
        return self.amount * float(NORM_RATIO[self.units])  # type: ignore


    def humanize (
        self,
        ) -> str:
        """
Adapted from:
<https://stackoverflow.com/a/56499010/1698443>
        """
        (years, remainder) = divmod(self.normalize(), NORM_RATIO[DurationUnits.YEAR.value])
        (months, remainder) = divmod(remainder, NORM_RATIO[DurationUnits.MONTH.value])
        (days, remainder) = divmod(remainder, NORM_RATIO[DurationUnits.DAY.value])
        (hours, remainder) = divmod(remainder, NORM_RATIO[DurationUnits.HOUR.value])
        (minutes, seconds) = divmod(remainder, NORM_RATIO[DurationUnits.MINUTE.value])

        cascade: zip = zip(
            NORM_RATIO.keys(),
            ( years, months, days, hours, minutes, seconds, ),
        )

        units: list = []

        for label, amount in cascade:
            if amount > 0:
                if amount > 1:
                    label = PLURAL.plural(label)

                units.append(f"{int(amount)} {label}")

        readable: str = ", ".join(units)
        return readable


    def get_model (
        self
        ) -> dict:
        """
Serializable representation for JSON.
        """
        return {
            "amount": self.amount,
            "units": self.units,
        }


class Temperature (Measure):  # pylint: disable=R0902
    """
A data class representing one parsed Temperature object.
    """

    @classmethod
    def build (
        cls,
        parse: typing.Any,
        ) -> "Temperature":
        """
Constructor from a `textx` parse object.
        """
        temperature_units: typing.Optional[ str ] = None

        if parse.units is not None:
            temperature_units = TemperatureUnits(parse.units).value

        return Temperature(
            amount = float(parse.amount),
            units = temperature_units,
        )


    @classmethod
    def celsius_to_fahrenheit (
        cls,
        amount: float,
        ) -> float:
        """
Convert from Celsius to Fahrenheit scale.
        """
        return (amount / 5.0 * 9.0) + 32.0


    @classmethod
    def fahrenheit_to_celsius (
        cls,
        amount: float,
        ) -> float:
        """
Convert from Fahrenheit to Celsius scale.
        """
        return (amount - 32.0) / 9.0 * 5.0


    def humanize (
        self
        ) -> str:
        """
HTML representation.
        """
        html: str = f"{self.amount} °{self.units}"

        if self.units == TemperatureUnits.CELSIUS.value:
            fahr: float = self.celsius_to_fahrenheit(self.amount)

            # round the converted temperature to the nearest 5 °F
            f_deg: int = int(round(fahr / 5.0) * 5.0)

            html += f" ({f_deg} °{TemperatureUnits.FAHRENHEIT.value})"

        return html



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Jupyter wrapper kernel for Bwyd
see copyright/license https://github.com/DerwenAI/bwyd/README.md
"""

from ipykernel.kernelapp import IPKernelApp  # pylint: disable=E0401
from ipykernel.kernelbase import Kernel  # pylint: disable=E0401


class BwydKernel (Kernel):  # pylint: disable=R0903
    """
Jupyter wrapper kernel
    """
    implementation = "Bwyd"
    implementation_version = "1.0"

    banner = "Bwyd kernel: a DSL for cooking"

    language = "no-op"
    language_version = "1.0"

    language_info = {
        "name": "Bwyd",
        "mimetype": "text/plain",
        "file_extension": ".bwyd",
    }


    def do_execute (  # type: ignore  # pylint: disable=R0913,R0917
        self,
        code,
        silent,
        store_history = True,  # pylint: disable=W0613
        user_expressions = None,  # pylint: disable=W0613
        allow_stdin = False,  # pylint: disable=W0613
        *,
        cell_meta = None,
        cell_id = None,
        ) -> dict:
        """
Simply echo any given input to `stdout`.
See: <https://ipython-books.github.io/16-creating-a-simple-kernel-for-jupyter/>
        """
        if not silent:
            self.send_response(
                self.iopub_socket,
                "stream",
                {
                    "name": "stdout",
                    "text": code,
                }
            )

        return {
            "status": "ok",
            "execution_count": self.execution_count, # base class increments execution count
            "payload": [],
            "user_expressions": {},
        }


    def do_apply (
        self,
        content,
        bufs,
        msg_id,
        reply_metadata,
        ):
        """DEPRECATED"""
        raise NotImplementedError


if __name__ == "__main__":
    IPKernelApp.launch_instance(kernel_class = BwydKernel)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
DSL implementing the Bwyd language.
see copyright/license https://github.com/DerwenAI/bwyd/README.md
"""

from collections import OrderedDict
from urllib.parse import ParseResult, urlparse
import datetime
import itertools
import json
import logging
import pathlib
import re
import typing

from icecream import ic  # type: ignore  # pylint: disable=E0401
from spdx_tools.spdx.validation.spdx_id_validators import is_valid_internal_spdx_id
import dateutil
import jinja2
import minify_html
import requests_cache
import spdx_license_list
import textx  # type: ignore  # pylint: disable=E0401

from .error import BwydParserError

from .measure import Converter, \
    Measure, DurationUnits, Duration, Temperature

from .ops import Dependency, \
    OpsTypes, OpNote, OpTransfer, OpAdd, OpAction, OpStore, OpHeat, OpChill, OpBake

from .resources import BWYD_SVG, JINJA_PAGE_TEMPLATE, URL_PATTERN

from .structure import Post, Product, \
    Activity, Focus, Closure


######################################################################
## module definitions

class Module:  # pylint: disable=R0902
    """
One parsed module.
    """
    def __init__ (
        self,
        path: pathlib.Path,
        parse_tree: typing.Any,
        converter: Converter,
        *,
        slug: typing.Optional[ str ] = None,
        ) -> None:
        """
Constructor.
        """
        self.path: pathlib.Path = path
        self.parse_tree: typing.Any = parse_tree
        self.converter: Converter = converter
        self.slug: typing.Optional[ str ] = slug
        self.title: str = ""
        self.text: str = ""
        self.cites: typing.List[ str ] = []
        self.posts: typing.List[ Post ] = []
        self.author: typing.Optional[ str ] = None
        self.spdx_id: typing.Optional[ str ] = None
        self.spdx_name: typing.Optional[ str ] = None
        self.updated: typing.Optional[ datetime.date ] = None
        self.closures: typing.Dict[ str, Closure ] = OrderedDict()


    @classmethod
    def _urlify (
        cls,
        text: str,
        ) -> str:
        """
Detect a URL within and format as HTML.
        """
        pat: str = r'<a target="_blank" href="\1">\1</a>'
        return URL_PATTERN.sub(pat, text)


    def get_image (
        self,
        ) -> str:
        """
Make the image URL embeddable in an <iframe/>
        """
        img_url: str = ""

        if len(self.posts) > 0:
            img_url = self.posts[0].get_image()

        return img_url


    def get_thumbnail (
        self,
        session: requests_cache.CachedSession,
        ) -> str:
        """
Accessor for a thumbnail URL.
        """
        img_url: str = BWYD_SVG

        if len(self.posts) > 0:
            img_url = self.posts[0].get_thumbnail(session)

        return img_url


    def get_schema_org (
        self,
        ) -> dict:
        """
Accessor for composing Schema.org metadata in JSON-LD
<https://schema.org/Recipe>
        """
        frag: dict = {
            "@context": "https://schema.org",
            "@type": "Recipe",
            "name": self.title,
            "description": self.text,
            "keywords": self.collect_keywords(),
            "recipeYield": self.total_yields()[0],
            "recipeIngredient": [
                measure.humanize_convert(entity.symbol, entity.external, self.converter) + " " + entity.text  # pylint: disable=C0301
                for entity, measure in self.iter_ingredients()
            ],
        }

        # optional metadata
        img_url: str = self.get_image()

        if len(img_url) > 0:
            frag["image"] = img_url

        if len(self.cites) > 0:
            frag["isBasedOn"] = self.cites[0]

        if self.updated is not None:
            frag["dateModified"] = self.updated.isoformat()

        if self.author is not None:
            author: str = self.author
            match: typing.Optional[ re.Match ] = re.match(r"^([\w\s]+).*$", author)

            if match is not None:
                author = match.group(1).strip()

            frag["author"] = author

        if self.spdx_id is not None:
            frag["license"] = f"https://spdx.org/licenses/{self.spdx_id}.html"

        return frag


    def get_model (
        self,
        ) -> dict:
        """
Return a list of JSON-friendly dictionary representations,
one for each parsed Closure.
        """
        closure_list: typing.List[ dict ] = [
            {
                "title": name,
                "yields": closure.total_yields(intermediaries = True),
                "text": closure.text,
                "supers": closure.supers,
                "keywords": closure.keywords,
                "requires": closure.get_dependencies(),
                "foci": [ focus.get_model(self.converter) for focus in closure.foci ],
            }
            for name, closure in self.closures.items()
        ]

        spdx_license: typing.Optional[ dict ] = None
        updated: typing.Optional[ str ] = None

        if self.spdx_id is not None:
            spdx_license = {
                "id": self.spdx_id,
                "name": self.spdx_name,
            }

        if self.updated is not None:
            updated = self.updated.isoformat()

        return {
            "path": self.path.name,
            "icon": BWYD_SVG,
            "title": self.title,
            "text": self.text,
            "license": spdx_license,
            "details": {
                "serves": self.total_yields(),
                "duration": self.total_duration(),
                "keywords": self.collect_keywords(),
                "author": self.author,
                "updated": updated,
            },
            "ingredients": [
                {
                    "amount": measure.humanize_convert(
                        entity.symbol,
                        entity.external,
                        self.converter,
                    ),
                    "text": entity.text,
                }
                for entity, measure in self.iter_ingredients()
            ],
            "sources": self.cites,
            "gallery": [ post.url for post in self.posts],
            "image": self.get_image(),

            "closures": closure_list,
            "schema_org": json.dumps(self.get_schema_org()),
        }


######################################################################
## validation

    def _validate_license (
        self,
        lic_parse: typing.Any,
        ) -> None:
        """
Helper method to validate an SPDX license identifier.
See: <https://spdx.org/licenses/>
        """
        spdx_id: str = lic_parse.spdx_id
        valid_ref: bool = is_valid_internal_spdx_id(f"SPDXRef-{spdx_id}")

        if not valid_ref or spdx_id not in spdx_license_list.LICENSES:
            loc: dict = textx.get_location(lic_parse)

            raise BwydParserError(
                f"unknown SPDX license ID `{spdx_id}` referenced at {loc}",
                symbol = spdx_id,
            )

        self.spdx_id = spdx_id
        self.spdx_name = spdx_license_list.LICENSES[spdx_id].name


    @classmethod
    def _validate_url (  # type: ignore  # pylint: disable=R1710
        cls,
        entity: typing.Any,
        ) -> str:
        """
Helper method to parse one URL.
        """
        try:
            loc: dict = textx.get_location(entity)
            url: str = entity.url
            result: ParseResult = urlparse(url)

            if result.scheme not in [ "http", "https" ]:
                raise BwydParserError(
                    f"badly formatted URL `{url}` referenced at {loc}",
                    symbol = url,
                )
            return url

        except Exception as ex:  # pylint: disable=W0718
            print(ex)


    def validate (
        self,
        ) -> None:
        """
Validate the forward references for one Bwyd module.
        """
        local_names: typing.Set[ str ] = {
            product.symbol
            for closure in self.closures.values()
            for product in closure.products
        }

        for closure in self.closures.values():
            # check for zero reference counts
            for name, entity in closure.containers.items():
                if entity.ref_count < 1:
                    print(f"Container defined but not used: {name}")
                    print(entity.loc)

            for name, entity in closure.tools.items():
                if entity.ref_count < 1:
                    print(f"Tool defined but not used: {name}")
                    print(entity.loc)

            for name, entity in closure.ingredients.items():
                if entity.ref_count < 1:
                    print(f"Ingredient defined but not used: {name}")
                    print(entity.loc)

            # check for references outside this module
            for symbol, entity in closure.ingredients.items():
                if entity.external and symbol not in local_names:
                    raise BwydParserError(
                        f"CLOSURE `{symbol}` used but not defined {entity.loc}",
                        symbol = symbol,
                    )


######################################################################
## parsing methods

    def _interpret_dependency (
        self,
        closure: Closure,
        depend_parse: typing.Any,
        *,
        debug: bool = False,
        ) -> None:
        """
Interpret and resolve each dependency: container, tool, ingredient, use.
        """
        depend_class_name: str = depend_parse.__class__.__name__

        if debug:
            #ic(dir(depend_parse))
            ic(depend_parse)
            ic(
                depend_class_name,
                depend_parse.symbol,
                depend_parse.text,
            )

        if depend_class_name == "Container":
            # forward reference, to be resolved during this parsing pass
            closure.containers[depend_parse.symbol] = Dependency(
                loc = textx.get_location(depend_parse),
                symbol = depend_parse.symbol,
                text = depend_parse.text,
            )

        elif depend_class_name == "Tool":
            # forward reference, to be resolved during this parsing pass
            closure.tools[depend_parse.symbol] = Dependency(
                loc = textx.get_location(depend_parse),
                symbol = depend_parse.symbol,
                text = depend_parse.text,
            )

        elif depend_class_name == "Ingredient":
            # forward reference, to be resolved during this parsing pass
            closure.ingredients[depend_parse.symbol] = Dependency(
                loc = textx.get_location(depend_parse),
                symbol = depend_parse.symbol,
                text = depend_parse.text,
            )

        elif depend_class_name == "Use":
            # external forward reference, to be resolved on a subsequent pass
            closure.ingredients[depend_parse.symbol] = Dependency(
                loc = textx.get_location(depend_parse),
                symbol = depend_parse.symbol,
                text = depend_parse.text,
                external = True,
            )


    def _interpret_op (  # pylint: disable=R0911,R0912,R0915
        self,
        closure: Closure,
        op_parse: typing.Any,
        *,
        debug: bool = False,
        ) -> typing.Optional[ OpsTypes ]:
        """
Interpret the steps within an activity.
        """
        op_class_name: str = op_parse.__class__.__name__

        if debug:
            ic(op_parse)

        if op_class_name == "Note":
            if debug:
                ic(
                    op_class_name,
                    op_parse.text,
                )

            return OpNote(
                loc = textx.get_location(op_parse),
                text = op_parse.text,
            )

        if op_class_name == "Transfer":
            if debug:
                ic(
                    op_class_name,
                    op_parse.symbol,
                )

            # resolve local reference
            entity: typing.Optional[ typing.Any ] = None

            if op_parse.symbol in closure.ingredients:
                entity = closure.ingredients[op_parse.symbol]
                entity.ref_count += 1
            else:
                loc: dict = textx.get_location(op_parse)

                raise BwydParserError(
                    f"INGREDIENT `{op_parse.symbol}` used but not defined {loc}",
                    symbol = op_parse.symbol,
                )

            return OpTransfer(
                loc = textx.get_location(op_parse),
                symbol = op_parse.symbol,
                entity = entity,
            )

        if op_class_name == "Add":
            # resolve local reference
            entity = None

            if op_parse.symbol in closure.ingredients:
                entity = closure.ingredients[op_parse.symbol]
                entity.ref_count += 1
            else:
                loc = textx.get_location(op_parse)

                raise BwydParserError(
                    f"INGREDIENT `{op_parse.symbol}` used but not defined {loc}",
                    symbol = op_parse.symbol,
                )

            measure: Measure = Measure.build(op_parse.measure)

            if debug:
                ic(
                    op_class_name,
                    op_parse.symbol,
                    op_parse.text,
                    measure,
                    entity,
                )

            return OpAdd(
                loc = textx.get_location(op_parse),
                symbol = op_parse.symbol,
                measure = measure,
                text = op_parse.text,
                entity = entity,
            )

        if op_class_name == "Action":
            # resolve local reference
            entity = None

            if op_parse.symbol in closure.tools:
                entity = closure.tools[op_parse.symbol]
                entity.ref_count += 1
            elif op_parse.symbol in closure.containers:
                entity = closure.containers[op_parse.symbol]
                entity.ref_count += 1
            else:
                loc = textx.get_location(op_parse)

                raise BwydParserError(
                    f"ACTION OBJECT `{op_parse.symbol}` used but not defined {loc}",
                    symbol = op_parse.symbol,
                )

            duration: Duration = Duration.build(op_parse.duration)

            if debug:
                ic(
                    op_class_name,
                    op_parse.symbol,
                    op_parse.modifier,
                    op_parse.until,
                    duration,
                    entity,
                )

            return OpAction(
                loc = textx.get_location(op_parse),
                tool = entity,
                modifier = op_parse.modifier,
                until = op_parse.until,
                duration = duration,
            )

        if op_class_name == "Bake":
            # resolve local reference
            if op_parse.symbol in closure.containers:
                entity = closure.containers[op_parse.symbol]
                entity.ref_count += 1
            else:
                loc = textx.get_location(op_parse)

                raise BwydParserError(
                    f"BAKE CONTAINER `{op_parse.symbol}` used but not defined {loc}",
                    symbol = op_parse.symbol,
                )

            duration = Duration.build(op_parse.duration)
            temperature = Temperature.build(op_parse.temperature)

            if debug:
                ic(
                    op_class_name,
                    op_parse.symbol,
                    op_parse.modifier,
                    op_parse.until,
                    duration,
                    temperature,
                )

            return OpBake(
                loc = textx.get_location(op_parse),
                mode = op_class_name,
                container = entity,
                modifier = op_parse.modifier,
                until = op_parse.until,
                duration = duration,
                temperature = temperature,
            )

        if op_class_name == "Heat":
            # resolve local reference
            if op_parse.symbol in closure.containers:
                entity = closure.containers[op_parse.symbol]
                entity.ref_count += 1
            else:
                loc = textx.get_location(op_parse)

                raise BwydParserError(
                    f"HEAT CONTAINER `{op_parse.symbol}` used but not defined {loc}",
                    symbol = op_parse.symbol,
                )

            duration = Duration.build(op_parse.duration)

            if debug:
                ic(
                    op_class_name,
                    op_parse.symbol,
                    op_parse.modifier,
                    op_parse.until,
                    duration,
                )

            return OpHeat(
                loc = textx.get_location(op_parse),
                container = entity,
                modifier = op_parse.modifier,
                until = op_parse.until,
                duration = duration,
            )

        if op_class_name == "Chill":
            # resolve local reference
            if op_parse.symbol in closure.containers:
                entity = closure.containers[op_parse.symbol]
                entity.ref_count += 1
            else:
                loc = textx.get_location(op_parse)

                raise BwydParserError(
                    f"CHILL CONTAINER `{op_parse.symbol}` used but not defined {loc}",
                    symbol = op_parse.symbol,
                )

            duration = Duration.build(op_parse.duration)

            if debug:
                ic(
                    op_class_name,
                    op_parse.symbol,
                    op_parse.modifier,
                    op_parse.until,
                    duration,
                )

            return OpChill(
                loc = textx.get_location(op_parse),
                container = entity,
                modifier = op_parse.modifier,
                until = op_parse.until,
                duration = duration,
            )

        if op_class_name == "Store":
            # resolve local reference
            if op_parse.symbol in closure.containers:
                entity = closure.containers[op_parse.symbol]
                entity.ref_count += 1
            else:
                loc = textx.get_location(op_parse)

                raise BwydParserError(
                    f"STORE CONTAINER `{op_parse.symbol}` used but not defined {loc}",
                    symbol = op_parse.symbol,
                )

            duration = Duration.build(op_parse.duration)

            if debug:
                ic(
                    op_class_name,
                    op_parse.symbol,
                    op_parse.modifier,
                    duration,
                )

            return OpStore(
                loc = textx.get_location(op_parse),
                container = entity,
                modifier = op_parse.modifier,
                duration = duration,
            )

        ## OTHERWISE, parse fails ...
        return None


    def _interpret_focus (
        self,
        closure: Closure,
        focus_parse: typing.Any,
        *,
        debug: bool = False,
        ) -> None:
        """
Interpret the activities within a focus.
        """
        if debug:
            ic(
                focus_parse,
                focus_parse.symbol,
            )

        # resolve local reference
        if focus_parse.symbol in closure.containers:
            entity: typing.Any = closure.containers[focus_parse.symbol]
            entity.ref_count += 1
        else:
            loc: dict = textx.get_location(focus_parse)

            raise BwydParserError(
                f"CONTAINER `{focus_parse.symbol}` used but not defined {loc}",
                symbol = focus_parse.symbol,
            )

        focus = Focus(
            container = entity,
        )

        closure.foci.append(focus)

        for activity in focus_parse.activities:
            if debug:
                ic(
                    activity,
                    activity.text,
                )

            act: Activity = Activity(
                text = activity.text,
            )

            focus.activities.append(act)

            for op_parse in activity.ops:
                op_obj: OpsTypes = self._interpret_op(  # type: ignore
                    closure,
                    op_parse,
                    debug = debug,
                )

                act.ops.append(op_obj)


    def _interpret_ratio (
        self,
        closure: Closure,
        ratio_parse: typing.Any,
        *,
        debug: bool = False,
        ) -> None:
        """
Interpret the components within a ratio.
        """
        if debug:
            ic(
                ratio_parse.name,
                [ (part.symbol, part.components) for part in ratio_parse.parts ],
            )

        for part in ratio_parse.parts:
            if len(part.components) < 1:
                # resolve local reference
                if part.symbol in closure.ingredients:
                    entity: typing.Any = closure.ingredients[part.symbol]
                    entity.ref_count += 1
                else:
                    loc: dict = textx.get_location(part)

                    raise BwydParserError(
                        f"RATIO part `{part.symbol}` used but not defined {loc}",
                        symbol = part.symbol,
                    )

            ## OHFUCK: store representation of this ratio


    def _interpret_closure (  # pylint: disable=R0912,R0915
        self,
        closure_parse: typing.Any,
        *,
        debug: bool = False,
        ) -> Closure:
        """
Helper method to interpret one Closure.
        """
        # ensure that "null" metadata values keep their semantics
        text: str = ""

        if closure_parse.text is not None and len(closure_parse.text) > 0:
            text = closure_parse.text

        closure: Closure = Closure(
            name = closure_parse.name,
            obj = closure_parse,
            text = text,
        )

        # handle taxonomy and keywords
        if closure_parse.supers is not None:
            for symbol in closure_parse.supers.ids:
                closure.supers.append(symbol)

            if debug:
                ic(closure.supers)

        if closure_parse.keywords is not None:
            for symbol in closure_parse.keywords.ids:
                closure.keywords.append(symbol)

            if debug:
                ic(closure.keywords)

        # interpret each product
        for prod_parse in closure_parse.prods:
            if debug:
                ic(
                    prod_parse.symbol,
                    prod_parse.measure.amount,
                    prod_parse.measure.units,
                    prod_parse.intermediate,
                )

            closure.products.append(
                Product(
                    loc = textx.get_location(prod_parse),
                    symbol = prod_parse.symbol,
                    amount = Measure.build(prod_parse.measure),
                    intermediate = (prod_parse.intermediate == "INTERMEDIATE"),
                )
            )

        # resolve each dependency
        for depend_parse in closure_parse.depend:
            self._interpret_dependency(
                closure,
                depend_parse,
                debug = debug,
            )

        # interpret each focus
        for focus_parse in closure_parse.foci:
            self._interpret_focus(
                closure,
                focus_parse,
                debug = debug,
            )

        # interpret the ratio specification, if any
        if closure_parse.ratio is not None:
            self._interpret_ratio(
                closure,
                closure_parse.ratio,
                debug = debug,
            )

        return closure


    def interpret (
        self,
        *,
        debug: bool = False,
        ) -> None:
        """
Interpret one Bwyd module.
        """
        self.title = self.parse_tree.title
        self.text = self.parse_tree.text

        # parse the optional metadata
        for meta_parse in self.parse_tree.meta:
            meta_class_name: str = meta_parse.__class__.__name__
            loc: dict = textx.get_location(meta_parse)

            if meta_class_name == "Author":
                self.author = self._urlify(meta_parse.name)

            elif meta_class_name == "License":
                if self.spdx_id is not None:
                    # do not allow multiple licenses
                    spdx_id: str = meta_parse.spdx_id

                    raise BwydParserError(
                        f"redundant SPDX license ID `{spdx_id}` referenced at {loc}",
                        symbol = spdx_id,
                    )

                self._validate_license(meta_parse)

            elif meta_class_name == "Updated":
                if self.updated is not None:
                    # do not allow multiple dates
                    updated: str = meta_parse.date

                    raise BwydParserError(
                        f"redundant dates `{updated}` referenced at {loc}",
                        symbol = updated,
                    )

                self.updated = dateutil.parser.parse(meta_parse.date).date()

            elif meta_class_name == "Cite":
                self.cites.append(self._validate_url(meta_parse))

            elif meta_class_name == "Post":
                self.posts.append(Post(url = self._validate_url(meta_parse)))

        # parse each `CLOSURE`
        for closure_parse in self.parse_tree.closures:
            self.closures[closure_parse.name] = self._interpret_closure(
                closure_parse,
                debug = debug,
            )

        # validate the resulting parsed module
        self.validate()


######################################################################
## aggregate measures

    def total_duration (
        self,
        ) -> str:
        """
Accessor for the total duration of one Bwyd module.
        """
        total_sec: int = int(sum([  # type: ignore  # pylint: disable=R1728
            op.get_duration().normalize()
            for closure in self.closures.values()
            for focus in closure.foci
            for activity in focus.activities
            for op in activity.ops
        ]))

        return Duration(
            amount = total_sec,
            units = DurationUnits.SECOND.value,
        ).humanize()


    def total_yields (
        self,
        ) -> typing.List[ str ]:
        """
Accessor for the total, non-intermediate yields of one Bwyd module.
        """
        return [
            product
            for closure in self.closures.values()
            for product in closure.total_yields()
        ]


    def iter_ingredients (
        self,
        ) -> typing.Iterator[typing.Tuple[ Dependency, Measure ]]:
        """
Iterator for the aggregate ingredients in one Bwyd module.
        """
        ing: OrderedDict = OrderedDict()

        for closure in self.closures.values():  # pylint: disable=R1702
            for focus in closure.foci:
                for activity in focus.activities:
                    for op in activity.ops:
                        if isinstance(op, OpAdd) and not op.entity.external:
                            measure: Measure = op.measure
                            name: str = op.entity.symbol

                            if name not in ing:
                                ing[name] = [
                                    op.entity,
                                    Measure(
                                        amount = measure.amount,
                                        units = measure.units,
                                    ),
                                ]
                            elif measure.units == ing[name][1].units:
                                ing[name][1].amount += measure.amount
                            else:
                                error_msg: str = f"wrong units for ingredient list: {measure.units}"
                                logging.error(error_msg)

        for entity, measure in ing.values():
            yield entity, measure


    def collect_keywords (
        self,
        ) -> typing.List[ str ]:
        """
Accessor for the collected keywords in one Bwyd module.
        """
        return sorted(
            list(
                itertools.chain.from_iterable([
                    [ *closure.supers, *closure.keywords ]
                    for closure in self.closures.values()
                ])
            )
        )


######################################################################
## Jinja2 template rendering

    def render_template (
        self,
        *,
        template: jinja2.Template = JINJA_PAGE_TEMPLATE,
        minify: bool = True,
        ) -> str:
        """
Load a Jinja2 template and render the data model as HTML,
which is by default minified.
        """
        html: str = template.render(
            module = self.get_model()
        )

        if minify:
            return minify_html.minify(  # pylint: disable=E1101
                html,
                keep_html_and_head_opening_tags = True,
                minify_css = True,
                minify_js = True,
                remove_processing_instructions = True,
            )

        return html



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Operations objects in the Bwyd language.
see copyright/license https://github.com/DerwenAI/bwyd/README.md
"""

from collections import OrderedDict
import enum
import typing

from pydantic import BaseModel, NonNegativeInt

from .measure import Converter, Measure, DurationUnits, Duration, Temperature


######################################################################
## dependencies

class Dependency (BaseModel):  # pylint: disable=R0902
    """
A data class representing one parsed dependency:
Ingredient, Tool, Container, etc.
    """
    loc: dict
    symbol: str
    text: str
    ref_count: NonNegativeInt = 0
    external: bool = False


    def get_model (
        self
        ) -> dict:
        """
Serializable representation for JSON.
        """
        return {
            "name": self.symbol,
            "text": self.text,
        }


class DependencyDict (OrderedDict):
    """
A dictionary of a specific class of dependencies, which also provides
a local namespace.
    """
    def get_model (
        self
        ) -> list:
        """
Serializable representation for JSON.
        """
        return [ dep.get_model() for dep in self.values() ]


######################################################################
## operations

class Appliance (enum.StrEnum):
    """
An enumeration class representing string literals for kinds of appliances.
    """
    GENERIC = enum.auto()
    STOVE = enum.auto()
    COOLER = enum.auto()
    OVEN = enum.auto()


class OpGeneric (BaseModel):  # pylint: disable=R0902
    """
A data class representing a generic operation.
    """
    loc: dict
    ref_count: NonNegativeInt = 0


    def get_duration (
        self,
        ) -> Duration:
        """
Stub: Total duration.
        """
        return Duration(
            amount = 0.0,
            units = DurationUnits.SECOND.value,
        )


class OpNote (OpGeneric):  # pylint: disable=R0902
    """
Represents a collapsable Note, inline *within* an Activity, from the
Author/Cook for other Cooks.
    """
    text: str


    def get_model (
        self
        ) -> dict:
        """
Serializable representation for JSON.
        """
        return {
            "note": {
                "text": self.text,
            }
        }


class OpTransfer (OpGeneric):  # pylint: disable=R0902
    """
Represents the action of a Cook to Transfer an intermediate into
a Container from another Focus, still *within* the same Closure.
    """
    symbol: str
    entity: Dependency


    def get_model (
        self,
        ) -> dict:
        """
Serializable representation for JSON.
        """
        return {
            "transfer": {
                "name": self.symbol,
            },
        }


class OpAdd (OpGeneric):  # pylint: disable=R0902
    """
Represents the action of a Cook to Add a measured amount of an
ingredient into a Container within an Activity.
    """
    symbol: str
    measure: Measure
    text: str
    entity: Dependency


    def get_model (
        self,
        converter: Converter,
        ) -> dict:
        """
Serializable representation for JSON.
        """
        amount: str = self.measure.humanize_convert(
            self.symbol,
            self.entity.external,
            converter,
        )

        return {
            "name": self.symbol,
            "amount": amount,
            "text": self.text,
        }


class OpAction (OpGeneric):  # pylint: disable=R0902
    """
Represents the action of a Cook using a Tool to perform part of an
Activity on the food within a specific Container.
    """
    tool: Dependency
    modifier: str
    until: str
    duration: Duration


    def get_duration (
        self,
        ) -> Duration:
        """
Duration of this operation.
        """
        return self.duration


    def get_model (
        self
        ) -> dict:
        """
Serializable representation for JSON.
        """
        return {
            "action": {
                "tool": self.tool.symbol,
                "verb": self.modifier,
                "text": self.until,
                "time": self.duration.humanize(),
            }
        }


class OpStore (OpGeneric):  # pylint: disable=R0902
    """
Represents the process of a Cook on a Container to store the yield of
a Closure for a specified time period.
    """
    container: Dependency
    modifier: str
    duration: Duration


    def get_model (
        self
        ) -> dict:
        """
Serializable representation for JSON.
        """
        return {
            "store": {
                "text": self.modifier,
                "upto": self.duration.humanize(),
            }
        }


class OpAppliance (OpGeneric):  # pylint: disable=R0902
    """
Represents the process of an Appliance operating on the food within
a specific Container as part of an Activity.
    """
    container: Dependency
    modifier: str
    until: str
    duration: Duration
    appliance: str = Appliance.GENERIC


    def get_duration (
        self,
        ) -> Duration:
        """
Duration of this operation.
        """
        return self.duration


class OpHeat (OpAppliance):  # pylint: disable=R0902
    """
Represents an Appliance: stove, range, hotplate, camp fire --
used to *heat* in different modes.
    """
    appliance: str = Appliance.STOVE


    def get_model (
        self
        ) -> dict:
        """
Serializable representation for JSON.
        """
        return {
            self.appliance: {
                "text": self.modifier,
                "until": self.until,
                "time": self.duration.humanize(),
            }
        }


class OpChill (OpHeat):  # pylint: disable=R0902
    """
Represents an Appliance: cooler --
used to *chill* in different modes.
    """
    appliance: str = Appliance.COOLER


class OpBake (OpAppliance):  # pylint: disable=R0902
    """
Represents an Appliance: oven --
used to *bake* in different modes.
    """
    mode: str
    temperature: Temperature
    appliance: str = Appliance.OVEN


    def get_model (
        self
        ) -> dict:
        """
Serializable representation for JSON.
        """
        return {
            self.appliance: {
                "text": self.modifier,
                "until": self.until,
                "time": self.duration.humanize(),
                "mode": self.mode.lower(),
                "temperature": self.temperature.humanize(),
            }
        }


OpsTypes = typing.Union[
    OpNote,
    OpTransfer,
    OpAdd,
    OpAction,
    OpStore,
    OpHeat,
    OpChill,
    OpBake,
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
CLI installer for the Bwyd kernel.
see copyright/license https://github.com/DerwenAI/bwyd/README.md
"""

from ipykernel.kernelapp import IPKernelApp  # pylint: disable=E0401

from .kernel import BwydKernel

IPKernelApp.launch_instance(
    kernel_class = BwydKernel,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Objects in the Bwyd language.
see copyright/license https://github.com/DerwenAI/bwyd/README.md
"""

from urllib.parse import urlparse
import base64
import io
import logging
import itertools
import typing

from PIL import Image
from pydantic import BaseModel, NonNegativeInt
from upath import UPath
import requests
import requests_cache

from .measure import Measure, Converter

from .ops import Dependency, DependencyDict, \
    OpsTypes, OpAdd


######################################################################
## gallery classes

class Post (BaseModel):  # pylint: disable=R0902
    """
A data class representing one Post object.
    """
    url: str


    def get_image (
        self,
        ) -> str:
        """
Accessor for an embeddable URL.
        """
        host: typing.Optional[ str ] = urlparse(self.url).hostname

        if host and host.endswith(".instagram.com"):
            embed: UPath = UPath(self.url) / "embed"
            return embed.as_posix()

        return self.url


    def thumbify (
        self,
        img_url: str,
        session: requests_cache.CachedSession,
        ) -> str:
        """
Access an image by URL, resize to thumbnail, convert to a data URL.
        """
        data_url: str = img_url

        try:
            req: requests_cache.CachedResponse = session.get(  # type: ignore
                img_url,
                timeout = 10,
                stream = True,
            )

            image: Image = Image.open(req.raw)  # type: ignore

            max_size: typing.Tuple[ int, int ] = (50, 50,)
            image.thumbnail(max_size)  # type: ignore

            buffered: io.BytesIO = io.BytesIO()
            image.save(buffered, format = "JPEG")  # type: ignore

            hex_data: str = base64.b64encode(buffered.getvalue()).decode("utf-8")
            data_url = f"data:image/jpeg;base64,{hex_data}"

        except requests.exceptions.Timeout as ex:  # pylint: disable=W0612
            error_msg: str = f"URL read timeout: {img_url}"
            logging.error(error_msg)

        return data_url


    def get_thumbnail (
        self,
        session: requests_cache.CachedSession,
        ) -> str:
        """
Accessor for a thumbnail URL.
        """
        host: typing.Optional[ str ] = urlparse(self.url).hostname

        if host and host.endswith(".instagram.com"):
            embed: UPath = UPath(self.url) / "media" / "?size=l"
            return self.thumbify(embed.as_posix(), session)

        return self.url


######################################################################
## yields classes

class Product (BaseModel):  # pylint: disable=R0902
    """
A data class representing one Product object.
    """
    loc: dict
    symbol: str
    amount: Measure
    intermediate: bool
    ref_count: NonNegativeInt = 0


######################################################################
## structural classes

class Activity (BaseModel):  # pylint: disable=R0902
    """
A data class representing one Activity object.
    """
    text: str
    ops: typing.List[ OpsTypes ] = []


    def get_model (
        self,
        converter: Converter,
        ) -> dict:
        """
Serializable representation for JSON.
        """
        dat: dict = {
            "title": self.text,
            "steps": [
                {
                    "ingredients": [
                        op.get_model(converter)
                        for op in self.ops
                        if isinstance(op, OpAdd)
                    ]
                }
            ]
        }

        for op in self.ops:
            if not isinstance(op, OpAdd):
                dat["steps"].append(op.get_model())

        return dat


class Focus (BaseModel):  # pylint: disable=R0902
    """
A data class representing a parsed Focus object.
    """
    container: Dependency
    activities: typing.List[ Activity ] = []


    def get_model (
        self,
        converter: Converter,
        ) -> dict:
        """
Serializable representation for JSON.
        """
        return {
            "container": self.container.symbol,
            "activities": [ act.get_model(converter) for act in self.activities ],
        }


class Closure (BaseModel, arbitrary_types_allowed = True):  # pylint: disable=R0902
    """
A data class representing one parsed Closure object.
    """
    name: str
    obj: typing.Any
    text: str = ""
    supers: typing.List[ str ] = []
    keywords: typing.List[ str ] = []
    containers: DependencyDict = DependencyDict()
    tools: DependencyDict = DependencyDict()
    ingredients: DependencyDict = DependencyDict()
    foci: typing.List[ Focus ] = []
    products: typing.List[ Product ] = []


    def get_dependencies (
        self
        ) -> list:
        """
Serialized representation in JSON for the containers and tools.
        """
        return [
            dep.get_model()
            for dep in itertools.chain(self.containers.values(), self.tools.values())
        ]


    def total_yields (
        self,
        *,
        intermediaries: bool = False,
        ) -> typing.List[ str ]:
        """
Accessor for the total, non-intermediate yields of one Closure object.
        """
        yields_list: typing.List[ str ] = []

        for product in self.products:
            if intermediaries or not product.intermediate:
                amount: str = product.amount.humanize_convert(
                    product.symbol,
                    False,
                    None,
                )

                html: str = f"{amount} {product.symbol}".replace("_", " ").strip()
                yields_list.append(html)

        return yields_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Jupyter kernel installer for Bwyd
see copyright/license https://github.com/DerwenAI/bwyd/README.md
"""

import argparse
import json
import os
import pathlib
import shutil
import sys
import typing

from jupyter_client.kernelspec import KernelSpecManager  # pylint: disable=E0401
from IPython.utils.tempdir import TemporaryDirectory  # pylint: disable=E0401

from .resources import ICON_PATH


KERNEL_JSON: dict = {
    "argv": [ sys.executable, "-m", "bwyd", "-f", "{connection_file}" ],
    "display_name": "Bwyd",
    "language": "bwyd",
    "codemirror_mode": "shell",
    "env": { "PS1": "$" },
}


def install_my_kernel_spec (
    *,
    user: bool = True,
    prefix: typing.Optional[ str ] = None,
    ) -> None:
    """
Generate a `kernel.json` kernel spec file, then use it to invoke
the Jupyter kernel installer.
    """
    with TemporaryDirectory() as tmp_dir:
        os.chmod(tmp_dir, 0o755) # Starts off as 700, not user readable
        spec_path: str = os.path.join(tmp_dir, "kernel.json")

        with open(spec_path, "w", encoding = "utf-8") as fp:  # pylint: disable=C0103
            json.dump(
                KERNEL_JSON,
                fp,
                sort_keys = True,
            )

        shutil.copyfile(
            ICON_PATH,
            pathlib.Path(tmp_dir) / ICON_PATH.name,
        )

        print("Installing IPython kernel spec for Bwyd")

        KernelSpecManager().install_kernel_spec(
            tmp_dir,
            "bwyd",
            user = user,
            prefix = prefix,
        )


def _is_root (
    ) -> bool:
    """
Check whether this is running as a `root` user?
    """
    try:
        return os.geteuid() == 0
    except AttributeError:
        return False # assume not an admin on non-Unix platforms


def main (
    argv = None,
    ) -> None:
    """
Main entry point for the kernel installer.
    """
    parser = argparse.ArgumentParser(
        description = "Install KernelSpec for the Bwyd kernel"
    )

    prefix_locations = parser.add_mutually_exclusive_group()

    prefix_locations.add_argument(
        "--user",
        help = "Install KernelSpec in user\'s home directory",
        action = "store_true",
    )

    prefix_locations.add_argument(
        "--sys-prefix",
        help = "Install KernelSpec in `sys.prefix` which is useful in conda / virtualenv",
        action = "store_true",
        dest = "sys_prefix",
    )

    prefix_locations.add_argument(
        "--prefix",
        help = "Install KernelSpec in this prefix",
        default = None
    )

    args = parser.parse_args(argv)
    user = False
    prefix = None

    if args.sys_prefix:
        prefix = sys.prefix
    elif args.prefix:
        prefix = args.prefix
    elif args.user or not _is_root():
        user = True

    install_my_kernel_spec(
        user = user,
        prefix = prefix,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package definitions for the Bwyd DSL.
see copyright/license https://github.com/DerwenAI/bwyd/README.md
"""

from .dsl import Bwyd, Corpus, Graph

from .error import BwydParserError

from .kernel import BwydKernel

from .measure import Conversion, Converter, Humanized, \
    MeasureUnits, Measure, \
    DurationUnits, Duration, \
    Temperature

from .module import Module

from .ops import Dependency, DependencyDict, Appliance, \
    OpsTypes, OpNote, OpTransfer, OpAdd, OpAction, OpStore, OpHeat, OpChill, OpBake

from .resources import BWYD_NAMESPACE, BWYD_SVG, \
    CONVERT_PATH, GRAMMAR_PATH, ICON_PATH, \
    JINJA_PAGE_TEMPLATE, JINJA_INDEX_TEMPLATE, \
    URL_PATTERN

from .structure import Post, Product, \
    Activity, Focus, Closure



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
DSL implementing the Bwyd language.
see copyright/license https://github.com/DerwenAI/bwyd/README.md
"""

import json
import logging
import pathlib
import tomllib
import typing
import urllib.parse

from icecream import ic  # type: ignore  # pylint: disable=E0401
from rdflib.namespace import DCTERMS, RDF, SKOS, XSD  # pylint: disable=W0611
import rdflib
import requests_cache
import textx  # type: ignore  # pylint: disable=E0401

from .measure import Conversion, Converter
from .module import Module
from .resources import BWYD_NAMESPACE, BWYD_SVG, \
    CONVERT_PATH, GRAMMAR_PATH, JINJA_INDEX_TEMPLATE


######################################################################
## knowledge graph management

class Graph:
    """
A knowledge graph based on a corpus of Bwyd modules.
    """
    def __init__ (
        self,
        ) -> None:
        """
Constructor.
        """
        self.lang: str = "en"
        self.prefix: str = "bwyd"
        self.ns_bwyd: rdflib.Namespace = rdflib.Namespace(BWYD_NAMESPACE)
        self.graph: rdflib.Graph = rdflib.Graph()

        nm: rdflib.namespace.NamespaceManager = self.graph.namespace_manager
        nm.bind(self.prefix, self.ns_bwyd)


    def compose_iri (
        self,
        names: typing.List[ str ],
        ) -> rdflib.URIRef:
        """
Compose an IRI in the Bwyd namespace.
        """
        urn: str = ":".join([ urllib.parse.quote_plus(name) for name in names ])
        return rdflib.URIRef(self.ns_bwyd + urn)


    def compose_iri_instance (
        self,
        inst_symbol: str,
        *,
        sub_symbol: typing.Optional[ str ] = None,
        ) -> rdflib.URIRef:
        """
Compose an IRI in the Bwyd namespace for an instance of a class.
        """
        if sub_symbol is not None:
            return rdflib.URIRef(self.ns_bwyd + inst_symbol + "#" + sub_symbol)

        return rdflib.URIRef(self.ns_bwyd + inst_symbol)


    def compose_iri_literal (
        self,
        literal: str,
        *,
        lang: str,
        ) -> rdflib.Literal:
        """
Compose an IRI in the Bwyd namespace for a literal.
        """
        return rdflib.Literal(
            literal,
            lang = lang,
        )


    def add_tuple (
        self,
        s_obj: rdflib.URIRef,
        p_obj: rdflib.URIRef,
        o_obj: rdflib.term.Identifier,
        ) -> None:
        """
Add one RDF tuple to the graph.
        """
        self.graph.add(( s_obj, p_obj, o_obj, ))


    def serialize (
        self,
        *,
        format: str = "turtle",  # pylint: disable=W0622
        ) -> str:
        """
Return the serialized graph int the given format.
        """
        return self.graph.serialize(
            format = format,
            base = BWYD_NAMESPACE,
        )


######################################################################
## corpus operations

class Corpus:  # pylint: disable=R0903
    """
A corpus of Bwyd modules.
    """

    def __init__ (
        self,
        config: dict,
        converter: Converter,
        *,
        lang: str = "en",
        ) -> None:
        """
Constructor.
        """
        logging.basicConfig(format="%(asctime)s %(message)s")

        self.config: dict = config
        self.converter: Converter = converter
        self.lang: str = lang


    def get_cache (
        self,
        *,
        cache_path: typing.Optional[ pathlib.Path ] = None,
        cache_expire: typing.Optional[ int ] = None,
        ) -> requests_cache.CachedSession:
        """
Build a URL request cache session, optionally loading any
previous serialized cache from disk.
        """
        if cache_path is None:
            cache_path = pathlib.Path(self.config["bwyd"]["cache_path"])

        if cache_expire is None:
            cache_expire = self.config["bwyd"]["cache_expire"]

        session: requests_cache.CachedSession = requests_cache.CachedSession(
            backend = requests_cache.SQLiteCache(cache_path),
        )

        session.settings.expire_after = cache_expire

        return session


    def iter_files (
        self,
        dir_path: pathlib.Path,
        *,
        glob: str = "*.bwyd",
        ) -> typing.Iterator[ pathlib.Path ]:
        """
Iterator for listing the Bwyd modules in a given directory.
        """
        for bwyd_path in dir_path.rglob(glob):
            # filter out checkpoint files, if any
            # WHERE DO THESE COME FROM?
            if not bwyd_path.stem.endswith("-checkpoint"):
                yield bwyd_path


    def render_html_files (
        self,
        dir_path: pathlib.Path,
        *,
        glob: str = "*.bwyd",
        suffix: str = ".html",
        debug: bool = False,
        ) -> typing.List[ Module ]:
        """
Traverse the given directory, rendering Bwyd scripts as HTML in place.
Return a count of the modules processed.
        """
        modules: typing.List[ Module ] = []
        dsl: Bwyd = Bwyd()

        for bwyd_path in self.iter_files(dir_path, glob = glob):
            slug: str = bwyd_path.stem

            if debug:
                ic(bwyd_path.name)

            # parse the Bwyd module
            module: Module = dsl.parse(
                bwyd_path,
                slug = slug,
            )

            # interpret the parsed module
            module.interpret(
                debug = debug,
            )

            modules.append(module)

            # render HTML using the Jinja2 template
            html_path: pathlib.Path = bwyd_path.with_suffix(suffix)

            with open(html_path, "w", encoding = "utf-8") as fp:
                fp.write(module.render_template())

        return modules


    def render_discovery (
        self,
        modules: typing.List[ Module ],
        index_path: pathlib.Path,
        ) -> None:
        """
Render an HTML index for search/discovery across a directory of recipes.
        """
        mod_data: dict = {
            "corpus": {
                "icon": BWYD_SVG,
                "modules": [
                    {
                        "slug": module.slug,
                        "thumb": module.get_thumbnail(self.get_cache()),
                        "title": module.title,
                        "text": module.text,
                        "serves": module.total_yields(),
                        "duration": module.total_duration(),
                        "updated": module.updated,
                        "keywords": module.collect_keywords(),
                    }
                    for module in modules
                ],
            },
        }

        html: str = JINJA_INDEX_TEMPLATE.render(mod_data)

        with open(index_path, "w", encoding = "utf-8") as fp:
            fp.write(html)


    def build_graph (  # pylint: disable=R0914
        self,
        modules: typing.List[ Module ],
        *,
        debug: bool = False,  # pylint: disable=W0613
        ) -> Graph:
        """
Build a knowledge graph from the modules in this corpus.
        """
        graph: Graph = Graph()

        class_closure: rdflib.URIRef = graph.compose_iri_instance("Closure")
        class_ingredient: rdflib.URIRef = graph.compose_iri_instance("Ingredient")
        class_product: rdflib.URIRef = graph.compose_iri_instance("Product")
        class_recipe: rdflib.URIRef = graph.compose_iri_instance("Recipe")
        pred_component_of: rdflib.URIRef = graph.compose_iri_instance("ComponentOf")
        pred_produced_by: rdflib.URIRef = graph.compose_iri_instance("ProducedBy")
        pred_uses_ingredient: rdflib.URIRef = graph.compose_iri_instance("UsesIngredient")

        for module in modules:
            module_iri: rdflib.URIRef = graph.compose_iri([ module.slug ])  # type: ignore

            graph.add_tuple(
                module_iri,
                RDF.type,
                class_recipe,
            )

            for closure_symbol, closure in module.closures.items():
                closure_iri: rdflib.URIRef = graph.compose_iri([ module.slug, closure_symbol ])  # type: ignore  # pylint: disable=C0301

                graph.add_tuple(
                    closure_iri,
                    RDF.type,
                    class_closure,
                )

                graph.add_tuple(
                    closure_iri,
                    pred_component_of,
                    module_iri,
                )

                graph.add_tuple(
                    closure_iri,
                    SKOS.prefLabel,
                    graph.compose_iri_literal(closure.name, lang = self.lang),
                )

                graph.add_tuple(
                    closure_iri,
                    DCTERMS.description,
                    graph.compose_iri_literal(closure.text, lang = self.lang),
                )

                for super_class in closure.supers:
                    super_iri: rdflib.URIRef = graph.compose_iri([ super_class ])

                    graph.add_tuple(
                        closure_iri,
                        RDF.type,
                        super_iri,
                    )

                    graph.add_tuple(
                        closure_iri,
                        SKOS.broader,
                        super_iri,
                    )

                for keyword in closure.keywords:
                    keyword_iri: rdflib.URIRef = graph.compose_iri_instance("Keyword", sub_symbol = keyword)  # pylint: disable=C0301

                    graph.add_tuple(
                        closure_iri,
                        SKOS.related,
                        keyword_iri,
                    )

                for product in closure.products:
                    product_iri: rdflib.URIRef = graph.compose_iri([ module.slug, product.symbol ])  # type: ignore  # pylint: disable=C0301

                    graph.add_tuple(
                        product_iri,
                        RDF.type,
                        class_product,
                    )

                    graph.add_tuple(
                        product_iri,
                        pred_produced_by,
                        closure_iri,
                    )

                for dependency in closure.ingredients.values():
                    if not dependency.external:
                        ingredient_iri: rdflib.URIRef = graph.compose_iri_instance("Ingredient", sub_symbol = dependency.symbol)  # pylint: disable=C0301

                        graph.add_tuple(
                            ingredient_iri,
                            RDF.type,
                            class_ingredient,
                        )

                        graph.add_tuple(
                            ingredient_iri,
                            SKOS.prefLabel,
                            graph.compose_iri_literal(dependency.symbol.replace("_", " "), lang = self.lang),  # pylint: disable=C0301
                        )

                        graph.add_tuple(
                            ingredient_iri,
                            DCTERMS.description,
                            graph.compose_iri_literal(dependency.text, lang = self.lang),
                        )
                    else:
                        ingredient_iri = graph.compose_iri([ module.slug, dependency.symbol ])  # type: ignore  # pylint: disable=C0301

                    graph.add_tuple(
                        closure_iri,
                        pred_uses_ingredient,
                        ingredient_iri,
                    )

        return graph


######################################################################
## parser/interpreter definitions

class Bwyd:  # pylint: disable=R0903
    """
Bwyd DSL parser/interpreter.
    """
    META_MODEL: textx.metamodel.TextXMetaModel = textx.metamodel_from_file(
        GRAMMAR_PATH,
        debug = False, # True
    )

    with open(CONVERT_PATH, "r", encoding = "utf-8") as fp:
        UNIT_CONVERTER: Converter = {
            conv.symbol: conv
            for row in json.load(fp)
            for conv in [ Conversion.model_validate(row) ]
        }


    def __init__ (
        self,
        *,
        config_path: typing.Optional[ pathlib.Path ] = None,
        converter: Converter = UNIT_CONVERTER,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = {}

        if config_path is not None:
            with open(config_path, mode = "rb") as fp:
                self.config = tomllib.load(fp)

        self.converter: Converter = converter


    def extend_converter (
        self,
        conversions: typing.List[ Conversion ],
        ) -> None:
        """
Extend the measurements unit converter by merging with provided conversions.
        """
        for conv in conversions:
            self.converter[ conv.symbol ] = conv


    def parse (
        self,
        path: pathlib.Path,
        *,
        slug: typing.Optional[ str ] = None,
        debug: bool = False,
        ) -> Module:
        """
Initialize a parser to load one Bywd module from a file.
        """
        return Module(
            path,
            self.META_MODEL.model_from_file(
                path,
                debug = debug,
            ),
            self.converter,
            slug = slug,
        )


    def build_corpus (
        self,
        ) -> Corpus:
        """
Factory for initializing a corpus of Bywd modules.
        """
        return Corpus(
            config = self.config,
            converter = self.converter,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Bwyd language exception handling.
see copyright/license https://github.com/DerwenAI/bwyd/README.md
"""


######################################################################
## exception handling

class BwydParserError (Exception):
    """
General parser exception for the Bwyd language.
    """

    def __init__ (
        self,
        *args,
        **kwargs,
        ) -> None:
        """
Custom exception which notes a Bwyd language discrepancy.
        """
        super().__init__(*args)
        self.symbol = kwargs.get("symbol")
